#include "actor.h"
static str state_str [] = {"nap", "for", "toi", "swa", "wat", "lea", "ent", "don", "vis"};
static void _find_slot(LandArray * actors, Actor * self);
static void _handle_path_state(Actor * self);
static void _start_path(Actor * self);
static bool _watch(Actor * self, Actor * a);
static bool _feed(Actor * self, Actor * a);
str actor_state_str(Actor * self) {
    return state_str [self->state2];
}
static void _find_slot(LandArray * actors, Actor * self) {
    int n = land_array_count(actors);
    for (int i = 1; i < n; i += 1) {
        if (land_array_get(actors, i) == NULL) {
            self->eid = i;
            land_array_replace_nth(actors, i, self);
            break;
        }
    }
    if (self->eid == 0) {
        self->eid = n;
        land_array_add(actors, self);
    }
}
Actor* actor_new(int cid, float x, float y, bool is_animal) {
    Actor * self;
    land_alloc(self);
    _find_slot(game->actors, self);
    self->cid = cid;
    self->cx = x;
    self->cy = y;
    self->is_animal = is_animal;
    self->toilet = 0;
    self->hungry = 0;
    self->thirsty = 0;
    self->trash = 0;
    self->tired = 0;
    self->need = 0;
    self->state_path = StatePathNone;
    self->state2 = StateEntering;
    return self;
}
void actor_delivery(int cid, float x, float y) {
    Actor * self;
    land_alloc(self);
    _find_slot(game->flyers, self);
    self->cid = DataZooplane;
    self->cx = x;
    self->cy = y - 30;
    self->need = cid;
    self->rx = x;
    self->ry = y;
    self->flying = 60;
}
void actor_explosion(float x, float y) {
    for (int i = 0; i < 10; i += 1) {
        Actor * self;
        land_alloc(self);
        _find_slot(game->flyers, self);
        self->cid = DataFragment;
        self->cx = x + land_rnd(0, 1);
        self->cy = y + land_rnd(0, 1);
        self->flying = 0;
        float a = land_rnd(0, 2 * pi);
        float r = land_rnd(0, 1) / 48.0;
        self->rx = cos(a) * r;
        self->ry = sin(a) * r;
        self->rz = 5;
        self->scale = land_rnd(.5, 1.2);
    }
}
void actor_del(Actor * self) {
    land_free(self);
}
Actor* actor_get(int aid) {
    return land_array_get_or_none(game->actors, aid);
}
int actor_dx(Actor * self) {
    int dx [] = {1, 0, - 1, 0};
    return dx [self->d];
}
int actor_dy(Actor * self) {
    int dy [] = {0, - 1, 0, 1};
    return dy [self->d];
}
void actor_move(Actor * self) {
    float s = 1 / 48.0;
    float backupx = self->cx;
    float backupy = self->cy;
    int ox = self->cx;
    int oy = self->cy;
    Tile * t = map_get(ox, oy);
    self->cx += actor_dx(self) * s;
    self->cy += actor_dy(self) * s;
    self->frame++;
    self->frame %= 32;
    int nx = self->cx;
    int ny = self->cy;
    if (nx != ox || ny != oy) {
        Map * m = game->map;
        bool turn = 0;
        Tile * nt = map_get(nx, ny);
        if (nt->h > t->h) {
            turn = 1;
        }
        else if (self->state2 != StateLeaving && self->state2 != StateEntering && (nx < m->ax || ny < m->ay || nx >= m->ax + m->aw || ny >= m->ay + m->ah)) {
            turn = 1;
        }
        else if (map_add_actor(nt, self->eid)) {
            map_remove_actor(t, self->eid);
        }
        else {
            turn = 1 /* can't fit */;
        }
        if (turn) {
            self->cx = backupx;
            self->cy = backupy;
            self->d += land_rand(1, 3);
            self->d %= 4;
        }
    }
}
void actor_wander(Actor * self) {
    if (self->flying > 0) {
        self->flying--;
        return ;
    }
    int t = game->ticks;
    if (t < self->state_ts) {
        if (self->state2 == StateForaging) {
            actor_move(self);
        }
    }
    else {
        self->state_ts = t + land_rand(60 * 3, 60 * 10);
        int r = land_rand(0, 3);
        self->d += land_rand(1, 3);
        self->d %= 4;
        if (r == 0) {
            self->state2 = StateNapping;
        }
        else {
            self->state2 = StateForaging;
            self->hungry++;
            if (self->hungry == 50) {
                tag_new(self->cx, self->cy, 1, 0, 0, 10 * 60, "hungry animal");
            }
            if (self->hungry == 80) {
                tag_new(self->cx, self->cy, 1, 0, 0, 10 * 60, "starving animal");
            }
            if (self->hungry >= 100) {
                self->left = 1;
            }
        }
        actor_kill(self);
    }
}
void actor_kill(Actor * self) {
    int killer = zoo_is_killer(self->cid);
    if (killer == 0) {
        return ;
    }
    for (int y = - 1; y < 2; y += 1) {
        for (int x = - 1; x < 2; x += 1) {
            int kx = self->cx + x;
            int ky = self->cy + y;
            Tile * t = map_get(kx, ky);
            for (int oi = 0; oi < 4; oi += 1) {
                Actor * a = actor_get(t->actors [oi]);
                if (a && a->cid != self->cid && a->cid != DataZookeeper) {
                    int prey = zoo_is_prey(a->cid);
                    if (prey < killer) {
                        a->left = 1;
                        if (! a->is_animal) {
                            zoo->visitors_killed [0]++;
                            tag_new(self->cx, self->cy, 1, 0, 0, 20 * 60, "visitor killed");
                        }
                        return ;
                    }
                }
            }
        }
    }
}
int get_dir(float x1, float y1, float x2, float y2) {
    //
    // A-----|
    //       -|
    //        X
    //
    float dx = fabs(x2 - x1);
    float dy = fabs(y2 - y1);
    if (dx > dy) {
        if (x1 < x2) {
            return 0;
        }
        else if (x1 > x2) {
            return 2;
        }
        else {
            return - 1;
        }
    }
    else {
        if (y1 < y2) {
            return 3;
        }
        else if (y1 > y2) {
            return 1;
        }
        else {
            return - 1;
        }
    }
}
void add_dir(int * x, int * y, int d) {
    if (d == 0) {
        (* x)++;
    }
    if (d == 1) {
        (* y)--;
    }
    if (d == 2) {
        (* x)--;
    }
    if (d == 3) {
        (* y)++;
    }
}
static void _handle_path_state(Actor * self) {
    if (self->state_path == StatePathWait) {
        if (game->ticks < self->state_ts) {
            if (game->ticks % 4 == 0) {
                self->d = land_rand(0, 3);
                actor_move(self);
            }
            else {
                self->state_path = StatePathCellReached;
            }
        }
        return ;
    }
    if (self->state_path == StatePathCellReached || self->state_path == StatePathWalking) {
        int ix = self->cx;
        int iy = self->cy;
        int d = get_dir(ix, iy, self->tx, self->ty);
        if (d >= 0) {
            self->d = d;
            actor_move(self);
            self->state_path = StatePathWalking;
            return ;
        }
        self->state_path = StatePathStart /* and continue below */;
    }
    if (self->state_path == StatePathStart) {
        if (self->on_path < self->found) {
            int d = self->path [self->on_path];
            self->tx = self->cx;
            self->ty = self->cy;
            add_dir(& self->tx, & self->ty, d);
            self->on_path++;
            self->state_path = StatePathCellReached;
        }
        else {
            if (self->on_path == self->found) {
                self->state_path = StatePathDone;
            }
            self->found = 0;
        }
    }
}
static void _start_path(Actor * self) {
    self->on_path = 0;
    self->state_path = StatePathStart;
}
void actor_visit(Actor * self) {
    _handle_path_state(self);
    int t = game->ticks;
    if (self->state_path == StatePathWalking || self->state_path == StatePathWait) {
        return ;
    }
    if (self->state_path == StatePathCellReached) {
        if (self->need == DataPortapot) {
            Tile * tile = map_get(self->tx, self->ty);
            if (tile->oid == DataPortapot) {
                tile->oid = DataPortapot + 1;
                self->state2 = StateToilet;
                self->state_ts = t + land_rand(60 * 5, 60 * 10);
                self->state_path = StatePathNone;
                self->need = 0;
            }
            else if (tile->oid == DataPortapot + 1) {
                self->state_path = StatePathWait;
                self->state_ts = t + 60 * 3;
            }
            return ;
        }
        else if (self->need) {
            Tile * tile = map_get(self->tx, self->ty);
            if (tile->oid == self->need) {
                self->state_ts = game->ticks + 30;
                int p = zoo_buy(self->need);
                if (p) {
                    zoo_budget(p);
                    tag_new(self->cx, self->cy, 0, 1, 0, 200, "+%d", p);
                }
                //print("visitor %d paid %d for %s",
                //    self.eid, p, data_name_for_id(game.data, self.need))
                if (tile->oid == DataVendingmachine) {
                    self->thirsty = 0;
                    if (land_probability(0.33)) {
                        self->trash = 1;
                    }
                }
                if (tile->oid == DataVendingmachineSnacks) {
                    self->hungry = 0;
                    int r = land_rand(0, 2);
                    if (r == 1) {
                        self->tired = 1;
                    }
                    if (r == 2) {
                        self->trash = 1;
                    }
                }
                if (tile->oid == DataBench) {
                    self->tired = 0;
                }
                if (tile->oid == DataTrashcan) {
                    self->trash = 0;
                }
                if (tile->oid == DataGiftshop) {
                    self->gift = 0;
                    self->gifted = 1;
                }
                self->state2 = StateNapping;
                self->state_path = StatePathNone;
                self->need = 0;
            }
        }
        else {
            return ;
        }
    }
    else if (self->state_path == StatePathDone) {
        self->state_path = StatePathNone;
        if (self->state2 == StateDone) {
            self->state2 = StateLeaving;
            actor_recap(self);
        }
        else if (self->visit < self->visits) {
            //print("%d/%d viewpoints visited", self.visit, self.visits)
            self->state2 = StateWatch;
            self->state_ts = game->ticks + 60 * 5;
            self->rx = floor(self->cx) + land_rnd(0.1, 0.9);
            self->ry = floor(self->cy) + land_rnd(0.1, 0.9);
        }
    }
    if (self->state2 == StateToilet) {
        if (t >= self->state_ts) {
            self->state2 = StateNapping;
            Tile * tile = map_get(self->tx, self->ty);
            if (tile->oid == DataPortapot + 1) {
                tile->oid = DataPortapot;
            }
            self->toilet = 0;
        }
        else {
            int ix = self->cx;
            int iy = self->cy;
            int d = get_dir(ix, iy, self->tx, self->ty);
            if (d >= 0) {
                self->d = d;
                actor_move(self);
            }
        }
        return ;
    }
    else if (t < self->state_ts) {
        if (self->state2 == StateWatch) {
            // todo: find closeby animal, turn towards it
            int d = get_dir(self->cx, self->cy, self->rx, self->ry);
            if (d >= 0) {
                if (d != self->d) {
                    if (self->turning > 0) {
                        self->turning--;
                    }
                    else {
                        self->d = d;
                        self->turning = 4;
                    }
                }
                actor_move(self);
            }
            if (fabs(self->cx - self->rx) + fabs(self->cy - self->ry) < 3 / 48.0) {
                self->state2 = StateWatching;
                actor_watch_animal(self);
                self->visit++;
                // 123456789
                //   T  T  T  T  T  T  T
                //     R    R    R    R
                //       H      H      H
                if (self->visit % 5 == 0) {
                    self->thirsty = 1;
                }
                if (self->visit % 11 == 0) {
                    self->toilet = 1;
                }
                if (self->visit % 17 == 0) {
                    self->hungry = 1;
                }
            }
        }
        else if (self->state2 == StateWatching) {
            ;
        }
    }
    else {
        if (self->state2 == StateEntering) {
            self->d = 2;
            actor_move(self);
            if ((int) self->cx == game->map->entrance_x) {
                self->state2 = StateNapping;
                int c = zoo_cost(DataGate);
                zoo_budget(c);
                tag_new(self->cx, self->cy, 0, 1, 0, 200, "+%d", c);
            }
        }
        else if (self->state2 == StateLeaving) {
            self->d = 0;
            actor_move(self);
            if ((int) self->cx >= game->map->entrance_x + 30) {
                self->left = 1;
            }
        }
        else if (self->state2 == StateDone) {
            self->found = path_find_position_on_road(self->cx, self->cy, game->map->entrance_x, game->map->entrance_y, 500, self->path);
            if (self->found == 0) {
                // FIXME... maybe go along the visited viewpoints in
                // reverse instead?
                print("FIXME: cannot find exit");
                self->state_ts += 60 * 10;
            }
            else {
                _start_path(self);
            }
        }
        else if (self->toilet) {
            self->state_ts = t + land_rand(60 * 3, 60 * 4);
            self->state2 = StateToilet;
            self->found = path_find_object_near_road(self->cx, self->cy, DataPortapot, 40, self->path);
            if (self->found == 0) {
                self->found = path_find_object_near_road(self->cx, self->cy, DataPortapot + 1, 25, self->path);
            }
            if (self->found == 0) {
                self->state2 = StateDone /* oh no, there's no toilet with 25 steps */;
                tag_new(self->cx, self->cy, 1, 0, 0, 10 * 60, "missing toilet");
            }
            else {
                _start_path(self);
                self->need = DataPortapot;
            }
        }
        else if (self->thirsty || self->hungry || self->tired || self->trash || self->gift) {
            int need = 0;
            int md = 25;
            if (self->thirsty) {
                need = DataVendingmachine;
            }
            else if (self->hungry) {
                need = DataVendingmachineSnacks;
                md = 50;
            }
            else if (self->tired) {
                need = DataBench;
            }
            else if (self->trash) {
                need = DataTrashcan;
            }
            else if (self->gift) {
                need = DataGiftshop;
                md = 75;
            }
            self->state_ts = t + land_rand(60 * 1, 60 * 2);
            self->state2 = StateForaging;
            self->found = path_find_object_near_road(self->cx, self->cy, need, md, self->path);
            if (self->found == 0) {
                //print("visitor %d could not find %s", self.eid,
                //    data_name_for_id(game.data, need))
                self->state2 = StateDone;
                if (self->thirsty) {
                    tag_new(self->cx, self->cy, 1, 0, 0, 10 * 60, "missing drinks");
                }
                else if (self->hungry) {
                    tag_new(self->cx, self->cy, 1, 0, 0, 10 * 60, "missing snacks");
                }
                else if (self->tired) {
                    tag_new(self->cx, self->cy, 1, 0, 0, 10 * 60, "exhausted");
                }
                else if (self->trash) {
                    tag_new(self->cx, self->cy, 1, 0, 0, 10 * 60, "trash");
                }
                else if (self->gift) {
                    ;
                }
            }
            else {
                //print("visitor %d found path of %d steps", self.eid, self.found)
                _start_path(self);
                self->need = need;
            }
        }
        else if (self->visit < self->visits) {
            self->found = path_find_position_on_road(self->cx, self->cy, self->visitx [self->visit], self->visity [self->visit], 400, self->path);
            if (self->found == 0) {
                print("visitor %d got stuck", self->eid);
                self->state2 = StateDone;
            }
            else {
                _start_path(self);
                self->state2 = StateVisit;
            }
        }
        else {
            if (self->gift || self->gifted) {
                // couldn't find one, or found one
                self->state2 = StateDone;
            }
            else {
                self->gift = 1;
                self->state2 = StateForaging;
            }
        }
    }
}
static bool _watch(Actor * self, Actor * a) {
    if (zoo_is_animal(a->cid)) {
        //print("visitor %d sees %s", self.eid,
        //    data_name_for_id(game.data, a.cid))
        self->visit_saw [self->visit] = a->cid;
        // only count each species once
        for (int vi = 0; vi < self->visit; vi += 1) {
            if (self->visit_saw [vi] == a->cid) {
                self->visit_saw [self->visit] = 0;
            }
        }
        self->animals_seen++;
        if (self->visit_saw [self->visit]) {
            self->species_seen++;
        }
        return 1;
    }
    return 0;
}
static bool _feed(Actor * self, Actor * a) {
    if (zoo_is_animal(a->cid)) {
        a->hungry = 0;
    }
    return 0;
}
void actor_watch_animal(Actor * self) {
    actor_radius_callback(self, 4, _watch);
}
void actor_feed_animals(Actor * self) {
    actor_radius_callback(self, 6, _feed);
}
void actor_radius_callback(Actor * self, int radius, bool(* cb)(Actor *, Actor *)) {
    // ccccccC
    // CbbbbBC
    // CBaaABC
    // CBA.ABC
    // CBAaaBC
    // CBbbbbC
    // Ccccccc
    int x = self->cx;
    int y = self->cy;
    for (int r = 1; r < radius; r += 1) {
        int l = r * 2;
        add_dir(& x, & y, 3) /* y+1 */;
        add_dir(& x, & y, 2) /* x-1 */;
        for (int d = 0; d < 4; d += 1) {
            for (int i = 0; i < l; i += 1) {
                add_dir(& x, & y, d);
                Tile * t = map_get(x, y);
                for (int oi = 0; oi < 4; oi += 1) {
                    Actor * a = actor_get(t->actors [oi]);
                    if (a) {
                        if (cb(self, a)) {
                            return ;
                        }
                    }
                }
            }
        }
    }
}
void actor_recap(Actor * self) {
    int plus = 0;
    if (self->toilet) {
        plus = (1 << game->level) * (- 10);
        print("visitor could not find a toilet, popularity %+d", plus);
    }
    else {
        plus = self->animals_seen * self->species_seen;
        str status = "";
        if (self->trash) {
            plus -= (1 << game->level);
            zoo->trash [0]++;
        }
        if (self->tired) {
            plus -= (1 << game->level);
            zoo->exhaustion [0]++;
        }
        if (self->thirsty) {
            plus /= 2;
            plus -= (1 << game->level) * 2;
            status = "thirsty ";
        }
        if (self->hungry) {
            plus /= 3;
            status = "hungry ";
            plus -= (1 << game->level) * 3;
        }
        if (self->thirsty && self->hungry) {
            status = "miserable ";
        }
        print("%svisitor %d saw %d animals and %d species, popularity %+d", status, self->eid, self->animals_seen, self->species_seen, plus);
    }
    game->popularity += plus;
    zoo_visitor_satisfaction(self->thirsty > 0, self->hungry > 0, self->toilet > 0, plus);
}
bool actor_to_cell(Actor * self, int nx, int ny) {
    Tile * tile2 = map_get(nx, ny);
    if (map_add_actor(tile2, self->eid)) {
        Tile * tile = map_get(self->cx, self->cy);
        map_remove_actor(tile, self->eid);
        self->cx = nx;
        self->cy = ny;
        return 1;
    }
    return 0;
}
void actor_keep(Actor * self) {
    _handle_path_state(self);
    if (self->state_path == StatePathWalking || self->state_path == StatePathWait) {
        return ;
    }
    if (self->state_path == StatePathCellReached) {
        return ;
    }
    int t = game->ticks;
    if (t < self->state_ts) {
        return ;
    }
    if (self->state2 == StateEntering) {
        self->home_x = self->cx;
        self->home_y = self->cy;
        self->state2 = StateNapping;
    }
    else if (self->state2 == StateNapping) {
        self->state2 = StateWatch;
        self->state_ts = t + 60 * 3;
    }
    else if (self->state2 == StateWatch) {
        Actor * a = zoo_get_hungry_animal(self->home_x, self->home_y, 25, 5);
        if (a) {
            int x = a->cx;
            int y = a->cy;
            for (int close = 1; close < 5; close += 1) {
                self->found = path_find_close_position_on_road_or_grass(self->cx, self->cy, x, y, 35, close, self->path);
                if (self->found > 0) {
                    break;
                }
            }
            if (self->found > 0) {
                self->state2 = StateForaging;
                _start_path(self);
            }
            else {
                self->state2 = StateNapping;
                print("unable to find path to %s", data_name_for_id(game->data, a->cid));
            }
        }
    }
    else if (self->state2 == StateForaging) {
        self->state2 = StateVisit;
        self->state_ts = t + 60 * 3;
    }
    else if (self->state2 == StateVisit) {
        if (self->state_path == StatePathDone) {
            self->state_path = StatePathNone;
            actor_feed_animals(self);
        }
        self->found = path_find_close_position_on_road_or_grass(self->cx, self->cy, self->home_x, self->home_y, 100, 1, self->path);
        if (self->found > 0) {
            self->state2 = StateNapping;
            _start_path(self);
        }
        else {
            self->state2 = StateForaging;
        }
    }
}
void actor_deliver(Actor * self) {
    self->frame++;
    self->frame %= 32;
    self->cy += 0.1;
    if (self->cy > self->ry) {
        if (self->need) {
            Actor * a = actor_new(self->need, self->rx, self->ry, 1);
            a->flying = self->flying;
            self->need = 0;
        }
        Tile * tile = map_get(self->cx, self->cy);
        if (tile->visible == 0) {
            self->left = 1;
        }
    }
}
void actor_explode(Actor * self) {
    if (fabs(self->rx) < fabs(self->ry)) {
        if (self->ry < 0) {
            self->frame++;
        }
        else {
            self->frame--;
        }
    }
    else {
        if (self->rx < 0) {
            self->frame--;
        }
        else {
            self->frame++;
        }
    }
    self->frame = (self->frame + 32) % 32;
    if (fabs(self->rx) < fabs(self->ry)) {
        self->frame += 32;
    }
    self->cx += self->rx;
    self->cy += self->ry;
    self->flying += self->rz;
    self->rz -= 0.15;
    if (self->flying < 0) {
        self->flying = 0;
        self->rz = - self->rz * 0.8;
        self->scale *= 0.6;
        if (self->rz < 1) {
            self->left = 1;
        }
    }
}
void actor_tick(Actor * self) {
    if (! self->in_map) {
        Tile * tile = map_get(self->cx, self->cy);
        if (! map_add_actor(tile, self->eid)) {
            return ;
        }
        self->in_map = 1;
    }
    if (self->is_animal) {
        actor_wander(self);
    }
    else if (self->cid == DataZookeeper) {
        actor_keep(self);
    }
    else {
        actor_visit(self);
    }
}
void actors_tick(void) {
    int n = land_array_count(game->actors);
    for (int i = 0; i < n; i += 1) {
        Actor * a = land_array_get(game->actors, i);
        if (a && a->left) {
            Tile * t = map_get(a->cx, a->cy);
            map_remove_actor(t, a->eid);
            actor_del(a);
            land_array_replace_nth(game->actors, i, NULL);
        }
    }
    int fn = land_array_count(game->flyers);
    for (int fi = 0; fi < fn; fi += 1) {
        Actor * a = land_array_get(game->flyers, fi);
        if (a && a->left) {
            actor_del(a);
            land_array_replace_nth(game->flyers, fi, NULL);
        }
    }
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(game->actors);
        for (Actor * a = LandArrayIterator_item(game->actors, &__iter0__); LandArrayIterator_next(game->actors, &__iter0__); a = LandArrayIterator_item(game->actors, &__iter0__)) {
            if (a) {
                actor_tick(a);
            }
        }
    }
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(game->flyers);
        for (Actor * a = LandArrayIterator_item(game->flyers, &__iter0__); LandArrayIterator_next(game->flyers, &__iter0__); a = LandArrayIterator_item(game->flyers, &__iter0__)) {
            if (a) {
                if (a->cid == DataZooplane) {
                    actor_deliver(a);
                }
                if (a->cid == DataFragment) {
                    actor_explode(a);
                }
            }
        }
    }
}
