#ifndef _ACTOR_
#define _ACTOR_
typedef struct Actor Actor;
#include "game.h"
#include "map.h"
#include "path.h"
enum {
    StateNapping=0,
    StateForaging=1,
    StateToilet=2,
    StateWatch=3,
    StateWatching=4,
    StateLeaving=5 /* walking out the door */,
    StateEntering=6,
    StateDone=7 /* looking for the exit */,
    StateVisit=8,
    StatePathNone=100,
    StatePathStart=101 /* a new path was just found, start walking it */,
    StatePathCellReached=102 /* just stepped onto a new cell */,
    StatePathWait=103 /* path is blocked, wait to re-try */,
    StatePathWalking=104 /* transitioning from one cell to the next */,
    StatePathDone=105 /* the destination was just reached */
};
struct Actor {
    int cid;
    int eid;
    float cx, cy /* cell position */;
    int frame;
    int d /* 0=1/0, 1=0/-1, 2=-1/0, 3=0/1 */;
    bool in_map;
    int state_path;
    int state2;
    int state_ts;
    bool is_animal;
    int path [500];
    int found;
    int on_path;
    int tx, ty;
    float rx, ry, rz;
    float scale;
    int flying;
    bool gifted;
    int hungry;
    int thirsty;
    int toilet;
    int tired;
    int trash;
    int gift;
    int need;
    int turning;
    int visit, visits;
    int visitx [100];
    int visity [100];
    int visit_saw [100] /* animal id (only one/first one) */;
    int species_seen;
    int animals_seen /* individuals, including same species */;
    bool left;
    int home_x, home_y;
};
str actor_state_str(Actor * self);
Actor* actor_new(int cid, float x, float y, bool is_animal);
void actor_delivery(int cid, float x, float y);
void actor_explosion(float x, float y);
void actor_del(Actor * self);
Actor* actor_get(int aid);
int actor_dx(Actor * self);
int actor_dy(Actor * self);
void actor_move(Actor * self);
void actor_wander(Actor * self);
void actor_kill(Actor * self);
int get_dir(float x1, float y1, float x2, float y2);
void add_dir(int * x, int * y, int d);
void actor_visit(Actor * self);
void actor_watch_animal(Actor * self);
void actor_feed_animals(Actor * self);
void actor_radius_callback(Actor * self, int radius, bool(* cb)(Actor *, Actor *));
void actor_recap(Actor * self);
bool actor_to_cell(Actor * self, int nx, int ny);
void actor_keep(Actor * self);
void actor_deliver(Actor * self);
void actor_explode(Actor * self);
void actor_tick(Actor * self);
void actors_tick(void);
#endif
