#include "draw.h"
static int _draw_pass;
void draw_cell(LandGrid * grid, LandView * view, int cell_x, int cell_y, float x, float y) {
    Tile * tile = map_get(cell_x, cell_y);
    float visible = map_visibility(cell_x, cell_y);
    float sy = view->scale_y;
    if (_draw_pass == PassGround) {
        // our tile is 64 pixels wide
        // the rendered tile 75% of the 256 pixel wide
        float s = 64 / (256 * 0.75);
        s *= 1.02 /* overlap slightly to avoid seams */;
        //if tile.draw_level == 0:
        //    land_image_draw_scaled(game.data.entry[DataGrass].image,
        //        x, y - (8 + tile.level * 20) * sy,
        //        s * view.scale_x, s * sy)
        if (visible >= 1 - epsilon) {
            land_image_draw_scaled(game->data->entry [tile->tid].image, x, y - (8 + tile->draw_level * 20) * sy, s * view->scale_x, s * sy);
        }
        else {
            land_image_draw_scaled_tinted(game->data->entry [tile->tid].image, x, y - (8 + tile->draw_level * 20) * sy, s * view->scale_x, s * sy, visible, visible, visible, 1);
        }
    }
    else if (_draw_pass == PassActors) {
        for (int i = 0; i < 4; i += 1) {
            Actor * a = actor_get(tile->actors [i]);
            if (a) {
                draw_actor(grid, view, a, tile);
            }
        }
        if (tile->oid && tile->oid != DataReserved) {
            float s = 64 / (256 * 0.75);
            if (visible >= 1 - epsilon) {
                land_image_draw_scaled(game->data->entry [tile->oid].image, x, y - (8 + tile->draw_level * 20) * sy, s * view->scale_x, s * sy);
            }
            else {
                land_image_draw_scaled_tinted(game->data->entry [tile->oid].image, x, y - (8 + tile->draw_level * 20) * sy, s * view->scale_x, s * sy, visible, visible, visible, 1);
            }
        }
    }
    else if (_draw_pass == PassGrid) {
        land_thickness(0);
        land_grid_isometric_placeholder(grid, view, cell_x, cell_y, x, y - tile->draw_level * 20 * sy);
        for (int i = 0; i < 4; i += 1) {
            Actor * a = actor_get(tile->actors [i]);
            if (! a) {
                continue;
            }
            float ax, ay;
            land_grid_get_cell_position(grid, view, a->cx, a->cy, & ax, & ay);
            ay -= tile->h * 20;
            land_color(0, 0, 1, 1);
            land_circle((ax - 5) * view->scale_x, (ay - 5) * view->scale_y, (ax + 5) * view->scale_x, (ay + 5) * view->scale_y);
            land_line(ax * view->scale_x, ay * view->scale_y, x, y - tile->draw_level * 20 * sy);
            land_font_set(small_font);
            land_text_pos((ax - 5) * view->scale_x, (ay - 5 - 20) * view->scale_y);
            land_print("%02d %s %d", tile->actors [i], actor_state_str(a), a->hungry);
        }
        if (tile->oid == DataReserved) {
            land_color(1, .5, 0, 1);
            land_line(x, y - 16 * view->scale_y, x, y + 16 * view->scale_y);
        }
    }
}
void draw_actor(LandGrid * grid, LandView * view, Actor * a, Tile * tile) {
    float s = 64 / (256 * 0.75);
    float x, y;
    land_grid_get_cell_position(grid, view, a->cx, a->cy, & x, & y);
    LandAnimation * anim = game->data->entry [a->cid].animation;
    int f = (int)(a->frame / 4);
    if (land_animation_length(anim) == 20) {
        if (f > 4) {
            f = 8 - f;
        }
        f += a->d * 5;
        y -= tile->h * 20;
    }
    y -= a->flying;
    if (a->cid == DataFragment) {
        s *= a->scale;
    }
    float yoffset = 30;
    LandImage * image = land_animation_get_frame(anim, f);
    float visible = tile->visible / 255.0f;
    if (visible >= 1 - epsilon) {
        land_image_draw_scaled(image, x * view->scale_x, y * view->scale_y - yoffset * view->scale_y, s * view->scale_x, s * view->scale_y);
    }
    else {
        land_image_draw_scaled_tinted(image, x * view->scale_x, y * view->scale_y - yoffset * view->scale_y, s * view->scale_x, s * view->scale_y, visible, visible, visible, 1);
    }
    if (game->debug_grid) {
        land_color_set(land_color_name("lime"));
        land_thickness(3);
        if (a->found) {
            float sx = view->scale_x;
            float sy = view->scale_y;
            float xp = x * sx;
            float yp = y * sy;
            for (int i = a->on_path; i < a->found; i += 1) {
                float xp2 = xp, yp2 = yp;
                int d = a->path [i];
                if (d == - 1) {
                    break;
                }
                if (d == 0) {
                    xp2 += 32 * sx;
                    yp2 += 16 * sy;
                }
                if (d == 1) {
                    xp2 += 32 * sx;
                    yp2 -= 16 * sy;
                }
                if (d == 2) {
                    xp2 -= 32 * sx;
                    yp2 -= 16 * sy;
                }
                if (d == 3) {
                    xp2 -= 32 * sx;
                    yp2 += 16 * sy;
                }
                land_line(xp, yp, xp2, yp2);
                xp = xp2;
                yp = yp2;
            }
        }
    }
}
void draw_tag(LandGrid * grid, LandView * view, Tag * tag) {
    float x, y;
    land_grid_get_cell_position(grid, view, tag->x, tag->y, & x, & y);
    land_color(tag->r, tag->g, tag->b, 1);
    land_text_pos(x * view->scale_x, (y - 48) * view->scale_x);
    land_print_center("%s", tag->text);
}
void draw_map(void) {
    _draw_pass = PassGround;
    land_map_draw(game->map->map, game->view);
    _draw_pass = PassActors;
    land_map_draw(game->map->map, game->view);
    _draw_pass = PassGrid;
    if (game->debug_grid) {
        land_map_draw(game->map->map, game->view);
    }
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(game->flyers);
        for (Actor * a = LandArrayIterator_item(game->flyers, &__iter0__); LandArrayIterator_next(game->flyers, &__iter0__); a = LandArrayIterator_item(game->flyers, &__iter0__)) {
            if (a) {
                draw_actor(game->map->grid, game->view, a, map_get(a->cx, a->cy));
            }
        }
    }
    land_font_set(small_font);
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(game->tags);
        for (Tag * tag = LandArrayIterator_item(game->tags, &__iter0__); LandArrayIterator_next(game->tags, &__iter0__); tag = LandArrayIterator_item(game->tags, &__iter0__)) {
            if (tag) {
                draw_tag(game->map->grid, game->view, tag);
            }
        }
    }
}
