#include "game.h"
Game * game;
static bool DEBUG = 0;
static bool _usable(int x, int y, int w, int h);
void game_init(void) {
    land_find_data_prefix("data/");
    land_alloc(game);
    game->data = data_load();
    game->view = land_view_new(0, 0, 1280, 720);
    game->actors = land_array_new();
    game->flyers = land_array_new();
    game->tags = land_array_new();
    zoo_init();
    game_new();
    strcpy(game->last_animal, "");
    strcpy(game->previous_animal, "");
    float scale = land_scale_to_fit(1280, 720, 256);
    data_reload_fonts(scale);
    menu_init();
}
void game_new(void) {
    if (game->map) {
        map_del(& game->map);
    }
    game->w = 200;
    game->h = 200;
    game->map = map_new(game->w, game->h);
    game->budget = 10000;
    game->popularity = 0;
    game->speed = 1;
    game->level = 0;
    game->male_lion = 1;
    land_array_clear(game->actors);
    land_array_add(game->actors, NULL);
    land_array_clear(game->flyers);
    land_array_add(game->flyers, NULL);
    land_array_clear(game->tags);
    land_array_add(game->tags, NULL);
    map_create();
    float cx = game->map->ax + game->map->aw / 2;
    float cy = game->map->ay + game->map->ah / 2;
    float x, y;
    land_grid_get_cell_position(game->map->map->first_layer->grid, NULL, cx, cy, & x, & y);
    land_view_scroll_center(game->view, x, y);
}
void game_tick(void) {
    float scale = land_scale_to_fit(1280, 720, 256);
    if (land_was_resized()) {
        data_reload_fonts(scale);
    }
    float xm = land_mouse_delta_x() * land_get_x_scale();
    float ym = land_mouse_delta_y() * land_get_y_scale();
    LandVector m = {land_mouse_x(), land_mouse_y(), 0};
    m = land_transform_vector(m);
    float mx = m.x;
    float my = m.y;
    game->mx = mx;
    game->my = my;
    int kx = 0, ky = 0;
    int click = 0;
    if (land_closebutton()) {
        land_quit();
    }
    if (land_key_pressed(LandKeyEscape)) {
        land_quit();
    }
    if (land_key(LandKeyLeft)) {
        kx--;
    }
    if (land_key(LandKeyRight)) {
        kx++;
    }
    if (land_key(LandKeyUp)) {
        ky--;
    }
    if (land_key(LandKeyDown)) {
        ky++;
    }
    if (land_key_pressed('n')) {
        game_new();
    }
    if (DEBUG) {
        if (land_key_pressed('9')) {
            game_new();
            map_add_debug_stuff();
        }
        if (land_key_pressed(LandKeyFunction + 2)) {
            game->popularity += 10000;
            game->budget += 100000;
        }
    }
    if (land_key_pressed('p')) {
        if (game->speed == 0) {
            game->speed = 1;
        }
        else {
            game->speed = 0;
        }
    }
    if (land_key_pressed('s')) {
        if (game->speed == 1) {
            game->speed = 10;
        }
        else {
            game->speed = 1;
        }
    }
    if (land_key_pressed(' ')) {
        menu_navigate("Zoo", 0);
    }
    if (land_key_pressed('d')) {
        game->debug_grid = ! game->debug_grid;
    }
    if (land_key_pressed('w')) {
        menu_navigate("Water", 0);
    }
    if (land_key_pressed('r')) {
        menu_navigate("Road", 0);
    }
    if (land_key_pressed('g')) {
        menu_navigate("Grass", 0);
    }
    if (land_key_pressed('e')) {
        menu_navigate("Enclosure", 0);
    }
    if (land_key_pressed('t')) {
        menu_navigate("Toilet", 0);
    }
    if (land_key_pressed('z')) {
        menu_navigate("Hire Zookeeper", 0);
    }
    if (land_key_pressed('b')) {
        menu_navigate("Amenities", 0);
    }
    if (land_key_pressed('a')) {
        menu_navigate("Buy Animals", 0);
    }
    if (land_key_pressed('i')) {
        menu_navigate("Info", 0);
    }
    if (land_key_pressed('m')) {
        menu_navigate("Manage", 0);
    }
    if (land_key_pressed('1')) {
        menu_navigate(game->last_animal, 1);
        click++;
    }
    if (land_key_pressed('2')) {
        menu_navigate(game->previous_animal, 1);
        click++;
    }
    if (land_key_pressed(LandKeyFunction + 12)) {
        land_screenshot(NULL);
    }
    if (kx || ky) {
        float s = 10 / game->view->scale_x;
        land_view_scroll(game->view, kx * s, 0);
        land_view_scroll(game->view, 0, ky * s);
    }
    if (land_mouse_delta_z()) {
        float z = (100 + land_mouse_delta_z()) / 100.0;
        land_view_scale(game->view, z, z);
    }
    if (land_mouse_button(LandButtonRight)) {
        float s = 1 / game->view->scale_x;
        land_view_scroll(game->view, - xm * s, - ym * s);
    }
    if (! menu_click(mx, my) && ! menu_shown()) {
        float sx, sy;
        map_get_under_mouse(mx, my, & sx, & sy);
        game->sx = sx;
        game->sy = sy;
        if (land_mouse_button_clicked(LandButtonLeft) || click) {
            Tile * tile = map_get(sx, sy);
            if (game->place == DataZookeeper) {
                if (_usable(sx - 1, sy - 1, 2, 2)) {
                    actor_new(game->place, sx, sy, 0);
                    zoo_budget(- zoo_cost(game->place));
                    game->place = 0;
                    map_place(sx, sy, FlagGrass, 2);
                    tile->oid = DataZookeeperHut;
                    map_get (sx - 1, sy)->oid = DataReserved;
                    map_get (sx - 1, sy - 1)->oid = DataReserved;
                    map_get (sx, sy - 1)->oid = DataReserved;
                }
            }
            else if (game->place == DataGiftshop) {
                if (_usable(sx - 1, sy - 1, 2, 2)) {
                    zoo_budget(- zoo_cost(game->place));
                    game->place = 0;
                    map_place(sx, sy, FlagGrass, 2);
                    tile->oid = DataGiftshop;
                    map_get (sx - 1, sy)->oid = DataReserved;
                    map_get (sx - 1, sy - 1)->oid = DataReserved;
                    map_get (sx, sy - 1)->oid = DataReserved;
                }
            }
            else if (zoo_is_animal(game->place)) {
                if (_usable(sx, sy, 1, 1)) {
                    int place = game->place;
                    if (place == DataLion) {
                        if (game->male_lion) {
                            place++;
                        }
                        game->male_lion = ! game->male_lion;
                    }
                    actor_delivery(place, sx, sy);
                    zoo_budget(- zoo_cost(game->place));
                    tag_new(sx, sy, 1, 1, 1, 200, "%s arriving", data_name_for_id(game->data, game->place));
                    game->place = 0;
                }
            }
            else if (game->place == DataPortapot || game->place == DataVendingmachine || game->place == DataVendingmachineSnacks || game->place == DataTrashcan || game->place == DataBench) {
                if (_usable(sx, sy, 1, 1)) {
                    tile->oid = game->place;
                    zoo_budget(- zoo_cost(game->place));
                    game->place = 0;
                }
            }
            else if (game->place == DataDirt) {
                if (_usable(sx - 2, sy - 2, 5, 5)) {
                    for (int y = - 2; y < 3; y += 1) {
                        for (int x = - 2; x < 3; x += 1) {
                            if (map_get (sx + x, sy + y)->draw_level > 0) {
                                actor_explosion(sx + x, sy + y);
                            }
                        }
                    }
                    map_enclosure(sx, sy, 5);
                    zoo_budget(- zoo_cost(game->place));
                    game->place = 0;
                }
            }
            else if (game->place == DataRoad) {
                if (_usable(sx - 1, sy - 1, 2, 2)) {
                    zoo_budget(- zoo_cost(game->place));
                    map_place(sx, sy, FlagRoad, 2);
                }
            }
            else if (game->place == DataGrass) {
                if (_usable(sx - 1, sy - 1, 2, 2) || tile->oid == DataZookeeperHut || tile->oid == DataGiftshop) {
                    map_harvest(sx, sy, 2);
                    zoo_budget(- zoo_cost(game->place));
                    map_place(sx, sy, FlagGrass, 2);
                }
            }
            else if (game->place == DataWater) {
                if (_usable(sx - 1, sy - 1, 2, 2)) {
                    zoo_budget(- zoo_cost(game->place));
                    map_place(sx, sy, FlagWater, 2);
                }
            }
        }
    }
    for (int i = 0; i < game->speed; i += 1) {
        game->ticks++;
        actors_tick();
        tags_tick();
        if (game->ticks % 60 == 0) {
            zoo_plan_visit();
        }
    }
}
static bool _usable(int x, int y, int w, int h) {
    return map_is_assigned(x, y, w, h) && ! map_is_reserved(x, y, w, h);
}
void game_draw(void) {
    land_unclip();
    land_scale_to_fit(1280, 720, 0);
    land_clear(0, 0, 0, 1);
    land_clip_transformed(0, 0, 1280, 720);
    draw_map();
    menu_draw();
    if (! menu_shown()) {
        ui_draw();
    }
}
void game_done(void) {
    ;
}
