#include "map.h"
#define _scramble_max(x, y) ((y) > (x) ? (y) : (x))
static Tile* _get(int x, int y);
static void _fixup(int x, int y, int base, int flag);
Tile* map_get(int x, int y) {
    return game->map->tiles + x + y * game->map->w;
}
void map_place_object(int x, int y, int oid) {
    Tile * t = map_get(x, y);
    t->oid = oid;
}
static Tile* _get(int x, int y) {
    return game->map->tiles + x + y * game->map->w;
}
int tile_count_actors(Tile * tile) {
    int c = 0;
    for (int i = 0; i < 4; i += 1) {
        if (tile->actors [i]) {
            c++;
        }
    }
    return c;
}
bool map_remove_actor(Tile * tile, int aid) {
    for (int i = 0; i < 4; i += 1) {
        if (tile->actors [i] == aid) {
            tile->actors [i] = 0;
            return 1;
        }
    }
    print("Error: cannot remove actor %d", aid);
    return 0;
}
bool map_add_actor(Tile * tile, int aid) {
    for (int i = 0; i < 4; i += 1) {
        if (tile->actors [i] == 0) {
            tile->actors [i] = aid;
            return 1;
        }
    }
    return 0;
}
static void _fixup(int x, int y, int base, int flag) {
    Tile * t0 = map_get(x, y);
    if (! (t0->flags & flag)) {
        return ;
    }
    int r = base;
    const int f1 = 128 + 1 + 2;
    const int f4 = 2 + 4 + 8;
    const int f16 = 8 + 16 + 32;
    const int f64 = 32 + 64 + 128;
    Tile * t [8]={map_get(x - 1, y - 1), map_get(x - 1, y - 0), map_get(x - 1, y + 1), map_get(x - 0, y + 1), map_get(x + 1, y + 1), map_get(x + 1, y - 0), map_get(x + 1, y - 1), map_get(x - 0, y - 1),};
    int m = 0;
    for (int i = 0; i < 8; i += 1) {
        if (t [i]->flags & flag) {
            m += 1 << i;
        }
    }
    if (m == 255) {
        r = base + 15;
    }
    else if (m == 255 - 1) {
        r = base + 9;
    }
    else if (m == 255 - 4) {
        r = base + 10;
    }
    else if (m == 255 - 16) {
        r = base + 11;
    }
    else if (m == 255 - 64) {
        r = base + 12;
    }
    else if (m == 255 - 1 - 16) {
        r = base + 13;
    }
    else if (m == 255 - 4 - 64) {
        r = base + 14;
    }
    else if ((m & (255 - 16 - 32 - 64)) == (255 - 16 - 32 - 64)) {
        r = base + 2;
    }
    else if ((m & (255 - 64 - 128 - 1)) == (255 - 64 - 128 - 1)) {
        r = base + 4;
    }
    else if ((m & (255 - 1 - 2 - 4)) == (255 - 1 - 2 - 4)) {
        r = base + 6;
    }
    else if ((m & (255 - 4 - 8 - 16)) == (255 - 4 - 8 - 16)) {
        r = base + 8;
    }
    else if ((m & f1) == f1) {
        r = base + 1;
    }
    else if ((m & f4) == f4) {
        r = base + 3;
    }
    else if ((m & f16) == f16) {
        r = base + 5;
    }
    else if ((m & f64) == f64) {
        r = base + 7;
    }
    t0->tid = r;
}
void fixup(int x, int y) {
    Tile * t0 = map_get(x, y);
    if (t0->flags & FlagWater) {
        _fixup(x, y, DataWater, FlagWater);
        t0->h = 1;
    }
    if (t0->flags & FlagDirt) {
        _fixup(x, y, DataDirt, FlagDirt);
        if (t0->tid >= DataDirt + 9) {
            t0->h = 0;
        }
        else {
            t0->h = 1;
        }
    }
    if (t0->flags & FlagRoad) {
        _fixup(x, y, DataRoad, FlagRoad);
        t0->h = 1;
    }
}
Map* map_new(int w, int h) {
    Map * self;
    land_alloc(self);
    self->tiles = land_calloc(w * h * sizeof (Tile));
    self->map = land_map_new();
    self->w = w;
    self->h = h;
    LandLayer * layer = land_layer_new();
    layer->view_h = 80 /* 2 full tile heights */;
    land_map_add_layer(self->map, layer);
    LandGrid * grid = land_isometric_custom_grid(32, 16, 32, 16, w, h, 0, draw_cell);
    land_layer_set_grid(layer, grid);
    self->grid = grid;
    return self;
}
void map_create(void) {
    int w = game->map->w;
    int h = game->map->h;
    int weights [] = {200, 50, 20, 1, 1, 1, 1, 50, 20, 1, 1, 1};
    for (int y = 0; y < h; y += 1) {
        for (int x = 0; x < w; x += 1) {
            int tid = land_select_random(weights, DataGrassLast - DataGrass + 1);
            Tile * t = _get(x, y);
            t->tid = DataGrass + tid;
            t->draw_level = 1;
            t->h = 1;
            int r = land_rand(0, 10);
            if (r == 2) {
                t->oid = DataTree1;
            }
            if (r == 3) {
                t->oid = DataTree2;
            }
        }
    }
    zoo_assign_for_level();
    game->map->entrance_x = 149;
    game->map->entrance_y = 138;
}
void map_add_debug_stuff(void) {
    game->budget += 10000;
    game->popularity += 1000;
    zoo_level_up();
    map_road(140, 137, 8, 2);
    map_road(140, 130, 2, 16);
    map_enclosure(140, 141, 5);
    actor_new(DataHippopotamus, 140, 141, 1);
    actor_new(DataHippopotamus, 140, 142, 1);
    map_enclosure(147, 141, 5);
    actor_new(DataPandabear, 147, 141, 1);
    actor_new(DataPandabear, 147, 142, 1);
    map_enclosure(147, 131, 5);
    actor_new(DataOpossum, 147, 131, 1);
    actor_new(DataOpossum, 147, 132, 1);
}
void map_del(Map * (* self_p)) {
    Map * self = * self_p;
    land_map_destroy(self->map);
    land_free(self->tiles);
    land_free(self);
    * self_p = NULL;
}
void map_get_under_mouse(float mx, float my, float * tx, float * ty) {
    mx /= game->view->scale_x;
    my /= game->view->scale_y;
    LandMap * m = game->map->map;
    LandLayer * l = m->first_layer;
    float s = 20;
    float x [3], y [3];
    Tile * t [3];
    land_grid_get_cell_at(l->grid, game->view, mx, my, x + 0, y + 0);
    land_grid_get_cell_at(l->grid, game->view, mx, my + s, x + 1, y + 1);
    //land_grid_get_cell_at(l.grid, game.view, mx, my - s * 2, x + 2, y + 2)
    int use = 0;
    t [0] = map_get(x [0], y [0]);
    t [1] = map_get(x [1], y [1]);
    if (t [0]->draw_level == 0) {
        if (t [1]->draw_level == 1) {
            use = 1;
        }
    }
    else if (t [0]->draw_level == 1 && t [1]->draw_level == 1) {
        use = 1;
    }
    //t[2] = map_get(x[2], y[2])
    * tx = x [use];
    * ty = y [use];
}
void map_place_water(int x, int y) {
    Tile * t = map_get(x, y);
    if (t) {
        t->flags |= FlagWater;
        for (int j = - 1; j < 2; j += 1) {
            for (int i = - 1; i < 2; i += 1) {
                fixup(x + i, y + j);
            }
        }
    }
}
void map_toggle(int x, int y, int flag) {
    Tile * t = map_get(x, y);
    if (t) {
        t->flags ^= flag;
        t->flags &= flag;
        if (! (t->flags & flag)) {
            t->tid = DataGrass;
        }
        t->draw_level = 1;
        if (t->flags & FlagDirt) {
            t->draw_level = 0;
        }
        for (int j = - 1; j < 2; j += 1) {
            for (int i = - 1; i < 2; i += 1) {
                fixup(x + i, y + j);
            }
        }
    }
}
void map_place(int x, int y, int flag, int s) {
    int sx = x - (int)(s / 2);
    int sy = y - (int)(s / 2);
    for (int j = sy; j < sy + s; j += 1) {
        for (int i = sx; i < sx + s; i += 1) {
            Tile * t = map_get(i, j);
            if (t) {
                if (flag == FlagGrass) {
                    t->flags = 0;
                    t->h = 1;
                    t->tid = DataGrass;
                }
                else {
                    t->flags = flag;
                }
                t->oid = 0;
                t->draw_level = 1;
            }
        }
    }
    for (int j = sy - 1; j < sy + s + 1; j += 1) {
        for (int i = sx - 1; i < sx + s + 1; i += 1) {
            fixup(i, j);
        }
    }
}
void map_harvest(int x, int y, int s) {
    int sx = x - (int)(s / 2);
    int sy = y - (int)(s / 2);
    for (int j = sy; j < sy + s; j += 1) {
        for (int i = sx; i < sx + s; i += 1) {
            Tile * t = map_get(i, j);
            if (t && (t->oid == DataTree1 || t->oid == DataTree2)) {
                t->oid = 0;
                zoo_budget(zoo_cost(DataTree1));
                tag_new(i, j, 0, 1, 0, 200, "+%d", zoo_cost(DataTree1));
            }
        }
    }
}
void map_enclosure(int x, int y, int s) {
    for (int j = y - (int)(s / 2); j < y - (int)(s / 2) + s; j += 1) {
        for (int i = x - (int)(s / 2); i < x - (int)(s / 2) + s; i += 1) {
            Tile * t = map_get(i, j);
            if (t) {
                t->flags = FlagDirt;
                t->oid = 0;
                t->draw_level = 0;
            }
        }
    }
    for (int j = y - (int)(s / 2) - 1; j < y - (int)(s / 2) - 1 + s + 2; j += 1) {
        for (int i = x - (int)(s / 2) - 1; i < x - (int)(s / 2) - 1 + s + 2; i += 1) {
            fixup(i, j);
        }
    }
}
float map_visibility(int x, int y) {
    Tile * t = map_get(x, y);
    return t->visible / 255.0f;
}
void map_assign(int w, int h) {
    int x = 150 - w;
    int y = 150 - h;
    game->map->ax = x;
    game->map->ay = y;
    game->map->aw = w;
    game->map->ah = h;
    for (int j = y - 25; j < y + h + 25; j += 1) {
        for (int i = x - 25; i < x + w + 25; i += 1) {
            Tile * t = map_get(i, j);
            int b = 0;
            if (i < x) {
                b = _scramble_max(x - i, b);
            }
            if (j < y) {
                b = _scramble_max(y - j, b);
            }
            if (i >= x + w) {
                b = _scramble_max(i + 1 - x - w, b);
            }
            if (j >= y + h) {
                b = _scramble_max(j + 1 - y - h, b);
            }
            t->visible = 255 - (int)(b * 255 / 25);
        }
    }
    x -= 1;
    y -= 1;
    w += 2;
    h += 2;
    for (int i = 1; i < w - 1; i += 1) {
        map_place_object(x + i, y, DataWall + 7);
    }
    for (int i = 1; i < w - 1; i += 1) {
        map_place_object(x + i, y + h - 1, DataWall + 3);
    }
    for (int i = 1; i < h - 1; i += 1) {
        map_place_object(x, y + i, DataWall + 1);
    }
    for (int i = 1; i < h - 1; i += 1) {
        map_place_object(x + w - 1, y + i, DataWall + 5);
    }
    map_place_object(x, y, DataWall + 0);
    map_place_object(x, y + h - 1, DataWall + 2);
    map_place_object(x + w - 1, y + h - 1, DataWall + 4);
    map_place_object(x + w - 1, y, DataWall + 6);
    map_road(148, 137, 52, 2);
    map_place_object(148 + 2, 137 + 1, DataGate);
}
bool map_is_reserved(int x, int y, int w, int h) {
    for (int j = y; j < y + h; j += 1) {
        for (int i = x; i < x + h; i += 1) {
            Tile * t = map_get(i, j);
            if (t->oid == DataReserved || t->oid == DataZookeeper || t->oid == DataGiftshop) {
                return 1;
            }
        }
    }
    return 0;
}
bool map_is_assigned(int x, int y, int w, int h) {
    return x >= game->map->ax && y >= game->map->ay && x + w <= game->map->ax + game->map->aw && y + h <= game->map->ay + game->map->ah;
}
void map_road(int x, int y, int w, int h) {
    for (int j = 0; j < h; j += 1) {
        for (int i = 0; i < w; i += 1) {
            Tile * t = map_get(x + i, y + j);
            if (t) {
                t->flags = FlagRoad;
                t->oid = 0;
            }
        }
    }
    for (int j = - 1; j < h + 1; j += 1) {
        for (int i = - 1; i < w + 1; i += 1) {
            fixup(x + i, y + j);
        }
    }
}
void tag_new(float x, float y, float r, float g, float b, int duration, str text, ...) {
    Tag * self;
    land_alloc(self);
    self->x = x;
    self->y = y;
    self->r = r;
    self->g = g;
    self->b = b;
    va_list args;
    va_start(args, text);
    vsprintf(self->text, text, args);
    va_end(args);
    self->until_ts = game->ticks + duration;
    land_array_add(game->tags, self);
}
void tags_tick(void) {
    int t = game->ticks;
    int i = 0, r = 0;
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(game->tags);
        for (Tag * tag = LandArrayIterator_item(game->tags, &__iter0__); LandArrayIterator_next(game->tags, &__iter0__); tag = LandArrayIterator_item(game->tags, &__iter0__)) {
            if (tag) {
                if (r == 0 && t > tag->until_ts) {
                    land_free(tag);
                    r = i;
                }
            }
            i++;
        }
    }
    if (r) {
        land_array_remove(game->tags, r);
    }
}
