#include "menu.h"
#define _scramble_max(x, y) ((y) > (x) ? (y) : (x))
static Menu _menu;
static void _cost(int i, int did);
static void _costv(int i, int cost);
static void _constrained(int i, int popularity);
static void _place(int did);
static void _buy(void);
static void _draw_status(void);
static void _draw_info(void);
static int _sum(int * a);
static void _graph(str name, int * a);
static void _draw_help(void);
void menu_init(void) {
    menu_items("Zoo", NULL);
    _menu.closed = 1;
    _menu.x = 8;
    _menu.y = 8;
}
void menu_items(str first, ...) {
    va_list args;
    va_start(args, first);
    int i = 0;
    while (1) {
        char const * x = first;
        if (i > 0) {
            x = va_arg(args, char const *);
        }
        if (! x) {
            break;
        }
        land_string_copy(_menu.items [i].name, x, 100);
        _menu.items [i].cost = 0;
        _menu.items [i].constrained = 0;
        i++;
    }
    va_end(args);
    _menu.n = i;
    _menu.buy = 0;
    _menu.closed = 0;
    _menu.x = game->mx;
    _menu.y = game->my;
    menu_measure();
}
static void _cost(int i, int did) {
    _menu.items [i].cost = zoo_cost(did);
}
static void _costv(int i, int cost) {
    _menu.items [i].cost = cost;
}
static void _constrained(int i, int popularity) {
    _menu.items [i].constrained = popularity;
}
static void _place(int did) {
    if (zoo_cost(did) > game->budget) {
        ;
    }
    else {
        game->place = did;
    }
}
static void _buy(void) {
    _menu.buy = 1;
    for (int i = 0; i < _menu.n; i += 1) {
        char * dname = land_lowercase_copy(_menu.items [i].name);
        int did = data_id_for_name(game->data, dname);
        land_free(dname);
        if (did) {
            _cost(i, did);
        }
    }
}
void menu_measure(void) {
    data_font_scale(1);
    land_font_set(medium_font);
    float h = land_line_height();
    _menu.w = 0;
    _menu.h = 0;
    for (int i = 0; i < _menu.n; i += 1) {
        float w = land_text_get_width(_menu.items [i].name);
        if (w > _menu.w) {
            _menu.w = w;
        }
        _menu.h += h;
    }
    if (_menu.x + _menu.w > 1280) {
        _menu.x = _scramble_max(0, 1280 - _menu.w);
    }
    if (_menu.y + _menu.h > 720) {
        _menu.y = _scramble_max(0, 720 - _menu.h);
    }
}
void menu_draw(void) {
    data_font_scale(1);
    land_font_set(medium_font);
    float h = land_line_height();
    float x = _menu.x;
    float y = _menu.y;
    land_color(0, 0, 0, 0.5);
    land_filled_rectangle(x, y, x + _menu.w + 8, y + _menu.h + 8);
    for (int i = 0; i < _menu.n; i += 1) {
        land_text_pos(x + 4, y + 4);
        if (_menu.items [i].constrained) {
            land_color(.5, .5, .5, 1);
        }
        else {
            land_color(1, 1, 1, 1);
        }
        land_print("%s", _menu.items [i].name);
        y += h;
    }
    int prices = 0;
    for (int i = 0; i < _menu.n; i += 1) {
        int c = _menu.items [i].cost;
        if (c > 0) {
            prices++;
        }
    }
    if (prices > 0) {
        x += _menu.w + 8;
        y = _menu.y;
        land_color(0, 0, 0, 0.5);
        land_filled_rectangle(x, y, x + 200, y + _menu.h + 8);
        for (int i = 0; i < _menu.n; i += 1) {
            land_text_pos(x + 4, y + 4);
            int pop = _menu.items [i].constrained;
            int c = 0;
            if (pop) {
                land_color(.5, .5, 1, 1);
                land_print("%d pop.", pop);
            }
            else {
                c = _menu.items [i].cost;
            }
            if (c > 0) {
                if (game->budget - c * 2 >= 0) {
                    land_color(0, .8, 0, 1);
                }
                else if (c <= game->budget) {
                    land_color(.8, .6, 0, 1);
                }
                else {
                    land_color(1, .1, .3, 1);
                }
                land_print("$%d", c);
            }
            y += h;
        }
    }
    _draw_status();
    if (_menu.show_info) {
        _draw_info();
    }
    if (_menu.show_help) {
        _draw_help();
    }
}
static void _draw_status(void) {
    data_font_scale(1);
    land_font_set(small_font);
    float x = 100;
    float y = 8;
    land_text_pos(x, y);
    land_color(0, 0, 0, 1);
    land_write("Level: %d • ", 1 + game->level);
    land_write("Budget: $%d • ", game->budget);
    land_write("Popularity: %d ", game->popularity);
}
static void _draw_info(void) {
    int w = 400;
    int h = 400;
    land_color(0, 0, 0, 0.5);
    land_filled_rectangle(640 - w / 2, 360 - h / 2, 640 - w / 2 + w, 360 - h / 2 + h);
    data_font_scale(1);
    land_color(1, 1, 1, 1);
    land_text_pos(640 - w / 2 + 8, 360 - h / 2 + 8);
    land_font_set(medium_font);
    land_print("Statistics");
    land_font_set(small_font);
    land_print("Max occupancy: %d", zoo->max_capacity);
    land_paragraph_start(0);
    land_write("Animals: %d • ", land_array_count(zoo->current_animals));
    land_write("Species: %d • ", land_hash_count(zoo->current_species_set));
    land_print("Viewpoints: %d", zoo->vp_count);
    land_paragraph_end();
    _graph("Sales", zoo->sales);
    _graph("Expenses", zoo->expenses);
    _graph("Popularity", zoo->popularity);
    _graph("Visitors", zoo->visitor_count);
    land_font_set(medium_font);
    land_print("Incidents");
    land_font_set(small_font);
    _graph("Hungry animals", zoo->hungry_count);
    _graph("No drinks", zoo->visitor_thirsty);
    _graph("No snacks", zoo->visitor_hungry);
    _graph("No toilet", zoo->visitor_toilet);
    _graph("Casualties", zoo->visitors_killed);
    _graph("Trash", zoo->trash);
    _graph("Exhaustion", zoo->exhaustion);
}
static int _sum(int * a) {
    int x = 0;
    for (int i = 0; i < DAYS; i += 1) {
        x += a [i];
    }
    return x;
}
static void _graph(str name, int * a) {
    land_write("%s: %d/%d", name, a [0], _sum(a));
    int bx = land_text_x();
    land_write(" ");
    int x = land_text_x_pos();
    int y = land_text_y_pos();
    int fh = land_line_height();
    int m = 0;
    land_color(1, 1, 1, 1);
    land_thickness(0);
    for (int i = 1; i < DAYS; i += 1) {
        if (a [i] > m) {
            m = a [i];
        }
    }
    if (m > 0) {
        int rh = fh - 4;
        for (int i = 1; i < DAYS - 1; i += 1) {
            land_line(x, y + 2 + rh - a [i] * rh / m, x + 1, y + 2 + rh - a [i + 1] * rh / m);
            x++;
        }
    }
    land_text_pos(bx, y + fh);
}
bool menu_click(float mx, float my) {
    if (! land_mouse_button_clicked(LandButtonLeft)) {
        return 0;
    }
    land_font_set(medium_font);
    float h = land_line_height();
    for (int i = 0; i < _menu.n; i += 1) {
        if (mx < _menu.x + _menu.w && my < _menu.y + (i + 1) * h) {
            menu_navigate(_menu.items [i].name, _menu.buy);
            return 1;
        }
    }
    menu_init();
    _menu.show_info = 0;
    _menu.show_help = 0;
    return 0;
}
static void _draw_help(void) {
    int w = 400;
    int h = 400;
    land_color(0, 0, 0, 0.5);
    land_filled_rectangle(640 - w / 2, 360 - h / 2, 640 - w / 2 + w, 360 - h / 2 + h);
    data_font_scale(1);
    land_color(.7, .8, 1, 1);
    land_text_pos(640 - w / 2 + 8, 360 - h / 2 + 8);
    land_font_set(medium_font);
    land_print("Keyboard shortcuts");
    land_font_set(small_font);
    land_print("SPACE");
    float sy = land_text_y();
    land_print("A");
    land_print("I");
    land_print("B");
    land_print("M");
    land_print("E");
    land_print("G");
    land_print("R");
    land_print("1");
    land_print("2");
    land_color(1, 1, 1, 1);
    land_text_pos(640 - w / 2 + 8 + 100, sy);
    land_print("Open menu");
    land_print("Buy Animals");
    land_print("Info");
    land_print("Build Amenities");
    land_print("Manage Zoo");
    land_print("Enclosure");
    land_print("Grass");
    land_print("Road");
    land_print("Repeat animal");
    land_print("Previous animal");
}
bool menu_shown(void) {
    return ! _menu.closed;
}
void menu_navigate(str nav_name, bool buy) {
    char t [100];
    land_string_copy(t, nav_name, 100) /* who knows what nav_name points to directly... */;
    if (land_equals(t, "Zoo")) {
        if (_menu.closed) {
            menu_items("Build", "Buy Animals", "Info", "Manage", "Settings", NULL);
        }
        else {
            menu_init();
        }
    }
    else if (land_equals(t, "Manage")) {
        menu_items("Hire Zookeeper", "Expand Area", "Start Expedition", NULL);
        _cost(0, DataZookeeper);
        _costv(1, (1 + game->level) * 1000);
        _costv(2, (1 + game->level) * 5000);
        int n = zoo_need_popularity();
        if (n) {
            _constrained(1, n);
        }
    }
    else if (land_equals(t, "Settings")) {
        menu_items("Help", "New game", "Exit game", NULL);
    }
    else if (land_equals(t, "Build")) {
        menu_items("Road", "Enclosure", "Water", "Grass", "Amenities", NULL);
        _cost(0, DataRoad);
        _cost(1, DataDirt);
        _cost(2, DataWater);
    }
    else if (land_equals(t, "Amenities")) {
        menu_items("Toilet", "Drinks", "Snacks", "Bench", "Trashcan", "Giftshop", NULL);
        _cost(0, DataPortapot);
        _cost(1, DataVendingmachine);
        _cost(2, DataVendingmachineSnacks);
        _cost(3, DataBench);
        _cost(4, DataTrashcan);
        _cost(5, DataGiftshop);
    }
    else if (land_equals(t, "Expand Area")) {
        if (zoo_level_up()) {
            menu_init();
        }
    }
    else if (land_equals(t, "Info")) {
        menu_init();
        _menu.show_info = ! _menu.show_info;
    }
    else if (land_equals(t, "Help")) {
        menu_init();
        _menu.show_help = 1;
    }
    else if (land_equals(t, "Buy Animals")) {
        menu_items("Badger", "Beaver", "Cat", "Cow", "Donkey", "Opossum", "Rat", "Sheep", "Wolf", "Exotic Animals", NULL);
        _buy();
    }
    else if (land_equals(t, "Exotic Animals")) {
        menu_items("Aardvark", "Anteater", "Chimpanzee", "Elephant", "Giraffe", "Hippopotamus", "Lion", "Pandabear", "Tiger", "Zebra", NULL);
        _buy();
    }
    else if (land_equals(t, "Enclosure")) {
        menu_init();
        _place(DataDirt);
    }
    else if (land_equals(t, "Road")) {
        menu_init();
        _place(DataRoad);
    }
    else if (land_equals(t, "Grass")) {
        menu_init();
        _place(DataGrass);
    }
    else if (land_equals(t, "Water")) {
        menu_init();
        _place(DataWater);
    }
    else if (land_equals(t, "Toilet")) {
        menu_init();
        _place(DataPortapot);
    }
    else if (land_equals(t, "Drinks")) {
        menu_init();
        _place(DataVendingmachine);
    }
    else if (land_equals(t, "Snacks")) {
        menu_init();
        _place(DataVendingmachineSnacks);
    }
    else if (land_equals(t, "Trashcan")) {
        menu_init();
        _place(DataTrashcan);
    }
    else if (land_equals(t, "Bench")) {
        menu_init();
        _place(DataBench);
    }
    else if (land_equals(t, "Giftshop")) {
        menu_init();
        _place(DataGiftshop);
    }
    else if (buy) {
        if (! land_equals(t, game->last_animal)) {
            land_string_copy(game->previous_animal, game->last_animal, 100);
        }
        land_string_copy(game->last_animal, t, 100);
        char * name = land_lowercase_copy(t);
        _place(data_id_for_name(game->data, name));
        land_free(name);
        menu_init();
        print("-> %s %s", game->last_animal, game->previous_animal);
    }
    else if (land_equals(t, "Hire Zookeeper")) {
        menu_init();
        _place(DataZookeeper);
    }
    else if (land_equals(t, "Exit game")) {
        menu_items("Never mind", "Quit!", NULL);
    }
    else if (land_equals(t, "Never mind")) {
        menu_init();
    }
    else if (land_equals(t, "Quit!")) {
        land_quit();
    }
    else if (land_equals(t, "New game")) {
        menu_items("Sure?", "New game!", NULL);
    }
    else if (land_equals(t, "Sure?")) {
        menu_init();
    }
    else if (land_equals(t, "New game!")) {
        menu_init();
        game_new();
    }
    else {
        print("unknown menu item %s", t);
    }
}
