#include "zoo.h"
#define _scramble_min(x, y) ((y) < (x) ? (y) : (x))
Zoo * zoo;
static void _animal(str name, int cost, int killer, int prey);
static void _calculate_level(void);
static void _shift(int * a);
static int _cmp_dist_order(void const * a, void const * b);
bool zoo_is_visitor(int aid) {
    if (aid == DataMom || aid == DataDad || aid == DataKid) {
        return 1;
    }
    return 0;
}
ZooAnimal* zoo_add_animal(str name, int did, int cost) {
    ZooAnimal * za;
    land_alloc(za);
    strncpy(za->name, name, sizeof (za->name));
    za->did = did;
    za->cost = cost;
    land_hash_insert(zoo->known_animals, name, za);
    return za;
}
static void _animal(str name, int cost, int killer, int prey) {
    int did = data_id_for_name(game->data, name);
    ZooAnimal * za = zoo_add_animal(name, did, cost * 100);
    za->killer = killer;
    za->prey = prey;
}
void zoo_init(void) {
    land_alloc(zoo);
    zoo->viewpoints = land_array_new();
    zoo->current_animals = land_array_new();
    zoo->known_animals = land_hash_new();
    zoo->current_species_set = land_hash_new();
    _animal("aardvark", 400, 0, 1);
    _animal("anteater", 400, 0, 1);
    _animal("badger", 10, 0, 1);
    _animal("beaver", 30, 0, 1);
    _animal("cat", 1, 1, 1);
    _animal("chimpanzee", 400, 0, 2);
    _animal("cow", 12, 0, 2);
    _animal("donkey", 20, 0, 2);
    _animal("elephant", 8000, 0, 3);
    _animal("giraffe", 2000, 0, 3);
    _animal("hippopotamus", 500, 0, 3);
    _animal("lion", 1000, 4, 4);
    _animal("lion_male", 1000, 4, 4);
    _animal("opossum", 8, 0, 0);
    _animal("pandabear", 10000, 0, 2);
    _animal("rat", 1, 0, 0);
    _animal("sheep", 5, 0, 1);
    _animal("tiger", 1000, 4, 4);
    _animal("wolf", 100, 3, 2);
    _animal("zebra", 500, 0, 2);
}
bool zoo_is_animal(int did) {
    str name = data_name_for_id(game->data, did);
    return land_hash_has(zoo->known_animals, name);
}
void zoo_plan_visit(void) {
    // get all viewing positions V in the zoo with a path next to an
    // enclosure
    // visitors:
    // type A (one time visitor): figure out a path to visit each single V
    // type B (regular): pick 10 random positions in V, visit those, and leave
    // type C (year pass holder): pick 1 position and visit just that
    //
    // - count the number of different species the visitor encounters
    //   (a bit random since the animals can move)
    // - visitors who are thirsty, hungry or can't find a toilet are unhappy
    //
    // ticket prices rise with popularity (?)
    // visitor number rises with popularity (?)
    //
    zoo->t++;
    if (zoo->t % 60 == 0) {
        _shift(zoo->species_count);
        _shift(zoo->animal_count);
        _shift(zoo->hungry_count);
        _shift(zoo->visitor_count);
        _shift(zoo->visitor_thirsty);
        _shift(zoo->visitor_hungry);
        _shift(zoo->visitor_toilet);
        _shift(zoo->visitors_killed);
        _shift(zoo->popularity);
        _shift(zoo->trash);
        _shift(zoo->exhaustion);
        _shift(zoo->died);
        _shift(zoo->expenses);
        _shift(zoo->sales);
    }
    zoo->vp_count = 0;
    int ax = game->map->ax;
    int ay = game->map->ay;
    int aw = game->map->aw;
    int ah = game->map->ah;
    for (int y = ay; y < ay + ah; y += 1) {
        for (int x = ax; x < ax + aw; x += 1) {
            Tile * t = map_get(x, y);
            if (t->flags & FlagRoad) {
                bool is_view = 0;
                for (int d = 0; d < 4; d += 1) {
                    int nx = x, ny = y;
                    add_dir(& nx, & ny, d);
                    Tile * nt = map_get(nx, ny);
                    if (nt->flags & FlagDirt) {
                        is_view = 1;
                    }
                }
                if (is_view) {
                    int ex = game->map->entrance_x;
                    int ey = game->map->entrance_y;
                    int steps [100];
                    int plen = path_find_position_on_road(ex, ey, x, y, 100, steps);
                    if (plen > 0) {
                        int i = zoo->vp_count++;
                        ZooViewpoint * vp;
                        if (i < land_array_count(zoo->viewpoints)) {
                            vp = land_array_get(zoo->viewpoints, i);
                        }
                        else {
                            land_alloc(vp);
                            land_array_add(zoo->viewpoints, vp);
                        }
                        vp->x = x;
                        vp->y = y;
                        vp->distance = plen;
                    }
                }
            }
        }
    }
    land_hash_clear(zoo->current_species_set);
    land_array_clear(zoo->current_animals);
    zoo->current_visitors = 0;
    zoo->hungry_count [0] = 0;
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(game->actors);
        for (Actor * a = LandArrayIterator_item(game->actors, &__iter0__); LandArrayIterator_next(game->actors, &__iter0__); a = LandArrayIterator_item(game->actors, &__iter0__)) {
            if (! a) {
                continue;
            }
            if (zoo_is_visitor(a->cid)) {
                if (a->state2 != StateLeaving) {
                    zoo->current_visitors++;
                }
            }
            else if (zoo_is_animal(a->cid)) {
                land_array_add(zoo->current_animals, a);
                str name = data_name_for_id(game->data, a->cid);
                if (! land_hash_has(zoo->current_species_set, name)) {
                    land_hash_insert(zoo->current_species_set, name, NULL);
                }
                if (a->hungry > 50) {
                    zoo->hungry_count [0]++;
                }
            }
        }
    }
    zoo->max_capacity = _scramble_min(_scramble_min(aw * ah / 25, land_array_count(zoo->current_animals)), zoo->vp_count);
    if (zoo->vp_count > 0 && land_array_count(zoo->current_animals) > 0 && zoo->current_visitors < zoo->max_capacity) {
        visitor_new();
    }
    _calculate_level();
}
static void _calculate_level(void) {
    int a = 200;
    int n = a;
    zoo->max_level = 0;
    zoo->need_popularity_for_levelup = n;
    for (int i = 0; i < game->level; i += 1) {
        if (game->popularity >= n) {
            zoo->max_level = 1 + i;
            a *= 2;
            n += a;
            zoo->need_popularity_for_levelup = n;
        }
        else {
            break;
        }
    }
}
int zoo_need_popularity(void) {
    if (game->popularity < zoo->need_popularity_for_levelup) {
        return zoo->need_popularity_for_levelup;
    }
    return 0;
}
static void _shift(int * a) {
    for (int i = DAYS - 1; i > 0; i += - 1) {
        a [i] = a [i - 1];
    }
    a [0] = 0;
}
void zoo_assign_for_level(void) {
    float fw = (1 + ((int)(game->level / 2))) * 25 / 2.0;
    float fh = fw + (game->level % 2) * 12.5;
    int w = ceil(fw);
    int h = ceil(fh);
    if (w > 100) {
        w = 100;
    }
    if (h > 100) {
        h = 100;
    }
    map_assign(w, h);
}
bool zoo_level_up(void) {
    int cost = (1 + game->level) * 1000;
    if (game->budget >= cost && game->popularity >= zoo->need_popularity_for_levelup) {
        zoo_budget(- cost);
        game->level++;
        zoo_assign_for_level();
        return 1;
    }
    return 0;
}
void visitor_new(void) {
    int visitors [] = {DataMom, DataDad, DataKid};
    int t = land_rand(0, 2);
    Actor * a = actor_new(visitors [t], game->map->entrance_x + 3, game->map->entrance_y, 0);
    int maxvp = 10;
    if (zoo->max_capacity > 10) {
        maxvp = land_rand(10, zoo->max_capacity);
    }
    a->visits = zoo_visit(maxvp, a->visitx, a->visity);
    a->visit = 0;
    a->state2 = StateEntering;
}
static int _cmp_dist_order(void const * a, void const * b) {
    ZooViewpoint * va = * (ZooViewpoint * const *) a;
    ZooViewpoint * vb = * (ZooViewpoint * const *) b;
    int ka = va->distance + va->order * 20;
    int kb = vb->distance + vb->order * 20;
    if (ka < kb) {
        return - 1;
    }
    if (ka > kb) {
        return + 1;
    }
    return 0;
}
int zoo_visit(int n, int * x, int * y) {
    if (zoo->vp_count == 0) {
        return 0;
    }
    LandArray * found = land_array_new();
    int last = zoo->vp_count - 1;
    for (int i = 0; i < n; i += 1) {
        int v = land_rand(0, last);
        ZooViewpoint * vp = land_array_get(zoo->viewpoints, v);
        land_array_swap(zoo->viewpoints, v, last);
        land_array_add(found, vp);
        vp->order = i;
        if (last == 0) {
            break;
        }
        last--;
    }
    land_array_sort(found, _cmp_dist_order);
    int i = 0;
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(found);
        for (ZooViewpoint * vp = LandArrayIterator_item(found, &__iter0__); LandArrayIterator_next(found, &__iter0__); vp = LandArrayIterator_item(found, &__iter0__)) {
            bool close = 0;
            for (int j = 0; j < i; j += 1) {
                int dx = x [i] - vp->x;
                int dy = y [i] - vp->y;
                if (abs(dx) + abs(dy) <= 3) {
                    close = 1;
                }
            }
            if (! close) {
                x [i] = vp->x;
                y [i] = vp->y;
                i++;
            }
        }
    }
    land_array_destroy(found);
    return i;
}
int zoo_cost(int did) {
    if (zoo_is_animal(did)) {
        str name = data_name_for_id(game->data, did);
        ZooAnimal * a = land_hash_get(zoo->known_animals, name);
        return a->cost;
    }
    if (did == DataZookeeper) {
        return 5000;
    }
    if (did == DataRoad) {
        return 10;
    }
    if (did == DataWater) {
        return 100;
    }
    if (did == DataDirt) {
        return 500;
    }
    if (did == DataVendingmachine) {
        return 1000;
    }
    if (did == DataVendingmachineSnacks) {
        return 2000;
    }
    if (did == DataPortapot) {
        return 4000;
    }
    if (did == DataTrashcan) {
        return 500;
    }
    if (did == DataBench) {
        return 1000;
    }
    if (did == DataGate) {
        int prices [] = {100, 200, 400, 800, 1500, 2500, 5000, 7500, 10000, 15000, 20000};
        return prices [game->level];
    }
    if (did == DataTree1) {
        return 1000;
    }
    if (did == DataGiftshop) {
        return 10000;
    }
    return 0;
}
int zoo_buy(int did) {
    if (did == DataVendingmachine) {
        return zoo_cost(DataGate) / 4;
    }
    if (did == DataVendingmachineSnacks) {
        return zoo_cost(DataGate);
    }
    if (did == DataTree1) {
        return zoo_cost(DataTree1);
    }
    if (did == DataGiftshop) {
        return zoo_cost(DataGate);
    }
    return 0;
}
int zoo_is_killer(int cid) {
    str name = data_name_for_id(game->data, cid);
    ZooAnimal * a = land_hash_get(zoo->known_animals, name);
    if (! a) {
        return 0;
    }
    return a->killer;
}
int zoo_is_prey(int cid) {
    str name = data_name_for_id(game->data, cid);
    ZooAnimal * a = land_hash_get(zoo->known_animals, name);
    if (! a) {
        return 2 /* includes visitors, zookeeper */;
    }
    return a->prey;
}
Actor* zoo_get_hungry_animal(int x, int y, int r, int shuffle) {
    Actor * found [shuffle];
    for (int i = 0; i < shuffle; i += 1) {
        found [i] = NULL;
    }
    int total = 0;
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(zoo->current_animals);
        for (Actor * a = LandArrayIterator_item(zoo->current_animals, &__iter0__); LandArrayIterator_next(zoo->current_animals, &__iter0__); a = LandArrayIterator_item(zoo->current_animals, &__iter0__)) {
            int dx = x - a->cx;
            int dy = y - a->cy;
            if (dx * dx + dy * dy > r * r) {
                continue;
            }
            bool ok = 0;
            Actor * least_hungry = NULL;
            for (int i = 0; i < shuffle; i += 1) {
                if (found [i] && (least_hungry == NULL || found [i]->hungry < least_hungry->hungry)) {
                    least_hungry = found [i];
                }
                if (found [i] == NULL) {
                    found [i] = a;
                    ok = 1;
                    total++;
                    break;
                }
            }
            if (! ok && least_hungry && a->hungry > least_hungry->hungry) {
                for (int i = 0; i < shuffle; i += 1) {
                    if (found [i] == least_hungry) {
                        found [i] = a;
                        break;
                    }
                }
            }
        }
    }
    if (total == 0) {
        return NULL;
    }
    int m = _scramble_min(shuffle - 1, total - 1);
    int rr = land_rand(0, m);
    return found [rr];
}
void zoo_visitor_satisfaction(bool thirsty, bool hungry, bool toilet, int popularity) {
    zoo->visitor_count [0]++;
    zoo->popularity [0] += popularity;
    if (thirsty) {
        zoo->visitor_thirsty [0]++;
    }
    if (hungry) {
        zoo->visitor_hungry [0]++;
    }
    if (toilet) {
        zoo->visitor_toilet [0]++;
    }
}
void zoo_budget(int diff) {
    game->budget += diff;
    if (diff > 0) {
        zoo->sales [0] += diff;
    }
    else {
        zoo->expenses [0] -= diff;
    }
}
